library angular2.src.change_detection.dynamic_change_detector_dart;

import 'package:angular2/src/facade/lang.dart'
    show isPresent, isBlank, BaseException, FunctionWrapper;
import 'package:angular2/src/facade/collection.dart'
    show List, ListWrapper, MapWrapper, StringMapWrapper;
import './parser/context_with_variable_bindings.dart'
    show ContextWithVariableBindings;
import './abstract_change_detector.dart' show AbstractChangeDetector;
import './change_detection_util.dart'
    show ChangeDetectionUtil, SimpleChange, uninitialized;
import './array_changes.dart' show ArrayChanges;
import './keyvalue_changes.dart' show KeyValueChanges;
import './proto_change_detector.dart'
    show
        ProtoRecord,
        RECORD_TYPE_SELF,
        RECORD_TYPE_PROPERTY,
        RECORD_TYPE_INVOKE_METHOD,
        RECORD_TYPE_CONST,
        RECORD_TYPE_INVOKE_CLOSURE,
        RECORD_TYPE_PRIMITIVE_OP,
        RECORD_TYPE_KEYED_ACCESS,
        RECORD_TYPE_INVOKE_FORMATTER,
        RECORD_TYPE_STRUCTURAL_CHECK,
        RECORD_TYPE_INTERPOLATE,
        ProtoChangeDetector;
import './interfaces.dart' show ChangeDetector, ChangeDispatcher;
import './exceptions.dart'
    show ExpressionChangedAfterItHasBeenChecked, ChangeDetectionError;

class DynamicChangeDetector extends AbstractChangeDetector {
  dynamic dispatcher;
  Map formatters;
  List values;
  List changes;
  List<ProtoRecord> protos;
  DynamicChangeDetector(
      dynamic dispatcher, Map formatters, List<ProtoRecord> protoRecords)
      : super() {
    this.dispatcher = dispatcher;
    this.formatters = formatters;
    this.values = ListWrapper.createFixedSize(protoRecords.length + 1);
    this.changes = ListWrapper.createFixedSize(protoRecords.length + 1);
    this.protos = protoRecords;
  }
  setContext(dynamic context) {
    ListWrapper.fill(this.values, uninitialized);
    this.values[0] = context;
  }
  detectChangesInRecords(bool throwOnChange) {
    List<ProtoRecord> protos = this.protos;
    var updatedRecords = null;
    for (var i = 0; i < protos.length; ++i) {
      ProtoRecord proto = protos[i];
      var change = this._check(proto);
      if (isPresent(change)) {
        var record =
            ChangeDetectionUtil.changeRecord(proto.bindingMemento, change);
        updatedRecords = ChangeDetectionUtil.addRecord(updatedRecords, record);
      }
      if (proto.lastInDirective && isPresent(updatedRecords)) {
        if (throwOnChange) ChangeDetectionUtil.throwOnChange(
            proto, updatedRecords[0]);
        this.dispatcher.onRecordChange(proto.directiveMemento, updatedRecords);
        updatedRecords = null;
      }
    }
  }
  _check(ProtoRecord proto) {
    try {
      if (proto.mode == RECORD_TYPE_STRUCTURAL_CHECK) {
        return this._structuralCheck(proto);
      } else {
        return this._referenceCheck(proto);
      }
    } catch (e) {
      throw new ChangeDetectionError(proto, e);
    }
  }
  _referenceCheck(ProtoRecord proto) {
    if (this._pureFuncAndArgsDidNotChange(proto)) {
      this._setChanged(proto, false);
      return null;
    }
    var prevValue = this._readSelf(proto);
    var currValue = this._calculateCurrValue(proto);
    if (!isSame(prevValue, currValue)) {
      this._writeSelf(proto, currValue);
      this._setChanged(proto, true);
      if (proto.lastInBinding) {
        return ChangeDetectionUtil.simpleChange(prevValue, currValue);
      } else {
        return null;
      }
    } else {
      this._setChanged(proto, false);
      return null;
    }
  }
  _calculateCurrValue(ProtoRecord proto) {
    switch (proto.mode) {
      case RECORD_TYPE_SELF:
        return this._readContext(proto);
      case RECORD_TYPE_CONST:
        return proto.funcOrValue;
      case RECORD_TYPE_PROPERTY:
        var context = this._readContext(proto);
        var c = ChangeDetectionUtil.findContext(proto.name, context);
        if (c is ContextWithVariableBindings) {
          return c.get(proto.name);
        } else {
          Function propertyGetter = proto.funcOrValue;
          return propertyGetter(c);
        }
        break;
      case RECORD_TYPE_INVOKE_METHOD:
        Function methodInvoker = proto.funcOrValue;
        return methodInvoker(this._readContext(proto), this._readArgs(proto));
      case RECORD_TYPE_KEYED_ACCESS:
        var arg = this._readArgs(proto)[0];
        return this._readContext(proto)[arg];
      case RECORD_TYPE_INVOKE_CLOSURE:
        return FunctionWrapper.apply(
            this._readContext(proto), this._readArgs(proto));
      case RECORD_TYPE_INTERPOLATE:
      case RECORD_TYPE_PRIMITIVE_OP:
        return FunctionWrapper.apply(proto.funcOrValue, this._readArgs(proto));
      case RECORD_TYPE_INVOKE_FORMATTER:
        var formatter = MapWrapper.get(this.formatters, proto.funcOrValue);
        return FunctionWrapper.apply(formatter, this._readArgs(proto));
      default:
        throw new BaseException('''Unknown operation ${proto.mode}''');
    }
  }
  _structuralCheck(ProtoRecord proto) {
    var self = this._readSelf(proto);
    var context = this._readContext(proto);
    var change = ChangeDetectionUtil.structuralCheck(self, context);
    if (isPresent(change)) {
      this._writeSelf(proto, change.currentValue);
    }
    return change;
  }
  _readContext(ProtoRecord proto) {
    return this.values[proto.contextIndex];
  }
  _readSelf(ProtoRecord proto) {
    return this.values[proto.selfIndex];
  }
  _writeSelf(ProtoRecord proto, value) {
    this.values[proto.selfIndex] = value;
  }
  _setChanged(ProtoRecord proto, bool value) {
    this.changes[proto.selfIndex] = value;
  }
  bool _pureFuncAndArgsDidNotChange(ProtoRecord proto) {
    return proto.isPureFunction() && !this._argsChanged(proto);
  }
  bool _argsChanged(ProtoRecord proto) {
    var args = proto.args;
    for (var i = 0; i < args.length; ++i) {
      if (this.changes[args[i]]) {
        return true;
      }
    }
    return false;
  }
  _readArgs(ProtoRecord proto) {
    var res = ListWrapper.createFixedSize(proto.args.length);
    var args = proto.args;
    for (var i = 0; i < args.length; ++i) {
      res[i] = this.values[args[i]];
    }
    return res;
  }
}
var _singleElementList = [null];
isSame(a, b) {
  if (identical(a, b)) return true;
  if (a is String && b is String && a == b) return true;
  if ((!identical(a, a)) && (!identical(b, b))) return true;
  return false;
}
