library angular2.src.change_detection.parser.ast_dart;

import 'package:angular2/src/facade/lang.dart'
    show
        FIELD,
        autoConvertAdd,
        isBlank,
        isPresent,
        FunctionWrapper,
        BaseException;
import 'package:angular2/src/facade/collection.dart'
    show List, Map, ListWrapper, StringMapWrapper;
import './context_with_variable_bindings.dart' show ContextWithVariableBindings;

class AST {
  eval(context) {
    throw new BaseException("Not supported");
  }
  get isAssignable {
    return false;
  }
  assign(context, value) {
    throw new BaseException("Not supported");
  }
  visit(visitor) {}
  String toString() {
    return "AST";
  }
}
class EmptyExpr extends AST {
  eval(context) {
    return null;
  }
  visit(visitor) {}
}
class Structural extends AST {
  AST value;
  Structural(AST value) : super() {
    this.value = value;
  }
  eval(context) {
    return value.eval(context);
  }
  visit(visitor) {
    return visitor.visitStructural(this);
  }
}
class ImplicitReceiver extends AST {
  eval(context) {
    return context;
  }
  visit(visitor) {
    return visitor.visitImplicitReceiver(this);
  }
}
class Chain extends AST {
  List expressions;
  Chain(List expressions) : super() {
    this.expressions = expressions;
  }
  eval(context) {
    var result;
    for (var i = 0; i < this.expressions.length; i++) {
      var last = this.expressions[i].eval(context);
      if (isPresent(last)) result = last;
    }
    return result;
  }
  visit(visitor) {
    return visitor.visitChain(this);
  }
}
class Conditional extends AST {
  AST condition;
  AST trueExp;
  AST falseExp;
  Conditional(AST condition, AST trueExp, AST falseExp) : super() {
    this.condition = condition;
    this.trueExp = trueExp;
    this.falseExp = falseExp;
  }
  eval(context) {
    if (this.condition.eval(context)) {
      return this.trueExp.eval(context);
    } else {
      return this.falseExp.eval(context);
    }
  }
  visit(visitor) {
    return visitor.visitConditional(this);
  }
}
class AccessMember extends AST {
  AST receiver;
  String name;
  Function getter;
  Function setter;
  AccessMember(AST receiver, String name, Function getter, Function setter)
      : super() {
    this.receiver = receiver;
    this.name = name;
    this.getter = getter;
    this.setter = setter;
  }
  eval(context) {
    var evaluatedContext = this.receiver.eval(context);
    while (evaluatedContext is ContextWithVariableBindings) {
      if (evaluatedContext.hasBinding(this.name)) {
        return evaluatedContext.get(this.name);
      }
      evaluatedContext = evaluatedContext.parent;
    }
    return this.getter(evaluatedContext);
  }
  get isAssignable {
    return true;
  }
  assign(context, value) {
    var evaluatedContext = this.receiver.eval(context);
    while (evaluatedContext is ContextWithVariableBindings) {
      if (evaluatedContext.hasBinding(this.name)) {
        throw new BaseException(
            '''Cannot reassign a variable binding ${this.name}''');
      }
      evaluatedContext = evaluatedContext.parent;
    }
    return this.setter(evaluatedContext, value);
  }
  visit(visitor) {
    return visitor.visitAccessMember(this);
  }
}
class KeyedAccess extends AST {
  AST obj;
  AST key;
  KeyedAccess(AST obj, AST key) : super() {
    this.obj = obj;
    this.key = key;
  }
  eval(context) {
    var obj = this.obj.eval(context);
    var key = this.key.eval(context);
    return obj[key];
  }
  get isAssignable {
    return true;
  }
  assign(context, value) {
    var obj = this.obj.eval(context);
    var key = this.key.eval(context);
    obj[key] = value;
    return value;
  }
  visit(visitor) {
    return visitor.visitKeyedAccess(this);
  }
}
class Formatter extends AST {
  AST exp;
  String name;
  List<AST> args;
  List<AST> allArgs;
  Formatter(AST exp, String name, List args) : super() {
    this.exp = exp;
    this.name = name;
    this.args = args;
    this.allArgs = ListWrapper.concat([exp], args);
  }
  visit(visitor) {
    return visitor.visitFormatter(this);
  }
}
class LiteralPrimitive extends AST {
  var value;
  LiteralPrimitive(value) : super() {
    this.value = value;
  }
  eval(context) {
    return this.value;
  }
  visit(visitor) {
    return visitor.visitLiteralPrimitive(this);
  }
}
class LiteralArray extends AST {
  List expressions;
  LiteralArray(List expressions) : super() {
    this.expressions = expressions;
  }
  eval(context) {
    return ListWrapper.map(this.expressions, (e) {
      return e.eval(context);
    });
  }
  visit(visitor) {
    return visitor.visitLiteralArray(this);
  }
}
class LiteralMap extends AST {
  List keys;
  List values;
  LiteralMap(List keys, List values) : super() {
    this.keys = keys;
    this.values = values;
  }
  eval(context) {
    var res = StringMapWrapper.create();
    for (var i = 0; i < this.keys.length; ++i) {
      StringMapWrapper.set(res, this.keys[i], this.values[i].eval(context));
    }
    return res;
  }
  visit(visitor) {
    return visitor.visitLiteralMap(this);
  }
}
class Interpolation extends AST {
  List strings;
  List expressions;
  Interpolation(List strings, List expressions) : super() {
    this.strings = strings;
    this.expressions = expressions;
  }
  eval(context) {
    throw new BaseException("evaluating an Interpolation is not supported");
  }
  visit(visitor) {
    visitor.visitInterpolation(this);
  }
}
class Binary extends AST {
  String operation;
  AST left;
  AST right;
  Binary(String operation, AST left, AST right) : super() {
    this.operation = operation;
    this.left = left;
    this.right = right;
  }
  eval(context) {
    var left = this.left.eval(context);
    switch (this.operation) {
      case '&&':
        return left && this.right.eval(context);
      case '||':
        return left || this.right.eval(context);
    }
    var right = this.right.eval(context);
    switch (this.operation) {
      case '+':
        return left + right;
      case '-':
        return left - right;
      case '*':
        return left * right;
      case '/':
        return left / right;
      case '%':
        return left % right;
      case '==':
        return left == right;
      case '!=':
        return left != right;
      case '<':
        return left < right;
      case '>':
        return left > right;
      case '<=':
        return left <= right;
      case '>=':
        return left >= right;
      case '^':
        return left ^ right;
      case '&':
        return left & right;
    }
    throw 'Internal error [\$operation] not handled';
  }
  visit(visitor) {
    return visitor.visitBinary(this);
  }
}
class PrefixNot extends AST {
  AST expression;
  PrefixNot(AST expression) : super() {
    this.expression = expression;
  }
  eval(context) {
    return !this.expression.eval(context);
  }
  visit(visitor) {
    return visitor.visitPrefixNot(this);
  }
}
class Assignment extends AST {
  AST target;
  AST value;
  Assignment(AST target, AST value) : super() {
    this.target = target;
    this.value = value;
  }
  eval(context) {
    return this.target.assign(context, this.value.eval(context));
  }
  visit(visitor) {
    return visitor.visitAssignment(this);
  }
}
class MethodCall extends AST {
  AST receiver;
  Function fn;
  List args;
  String name;
  MethodCall(AST receiver, String name, Function fn, List args) : super() {
    this.receiver = receiver;
    this.fn = fn;
    this.args = args;
    this.name = name;
  }
  eval(context) {
    var evaluatedContext = this.receiver.eval(context);
    var evaluatedArgs = evalList(context, this.args);
    while (evaluatedContext is ContextWithVariableBindings) {
      if (evaluatedContext.hasBinding(this.name)) {
        var fn = evaluatedContext.get(this.name);
        return FunctionWrapper.apply(fn, evaluatedArgs);
      }
      evaluatedContext = evaluatedContext.parent;
    }
    return this.fn(evaluatedContext, evaluatedArgs);
  }
  visit(visitor) {
    return visitor.visitMethodCall(this);
  }
}
class FunctionCall extends AST {
  AST target;
  List args;
  FunctionCall(AST target, List args) : super() {
    this.target = target;
    this.args = args;
  }
  eval(context) {
    var obj = this.target.eval(context);
    if (!(obj is Function)) {
      throw new BaseException('''${obj} is not a function''');
    }
    return FunctionWrapper.apply(obj, evalList(context, this.args));
  }
  visit(visitor) {
    return visitor.visitFunctionCall(this);
  }
}
class ASTWithSource extends AST {
  AST ast;
  String source;
  String location;
  ASTWithSource(AST ast, String source, String location) : super() {
    this.source = source;
    this.location = location;
    this.ast = ast;
  }
  eval(context) {
    return this.ast.eval(context);
  }
  get isAssignable {
    return this.ast.isAssignable;
  }
  assign(context, value) {
    return this.ast.assign(context, value);
  }
  visit(visitor) {
    return this.ast.visit(visitor);
  }
  String toString() {
    return '''${this.source} in ${this.location}''';
  }
}
class TemplateBinding {
  String key;
  bool keyIsVar;
  String name;
  ASTWithSource expression;
  TemplateBinding(
      String key, bool keyIsVar, String name, ASTWithSource expression)
      : super() {
    this.key = key;
    this.keyIsVar = keyIsVar;
    this.name = name;
    this.expression = expression;
  }
}
class AstVisitor {
  visitAccessMember(AccessMember ast) {}
  visitAssignment(Assignment ast) {}
  visitBinary(Binary ast) {}
  visitChain(Chain ast) {}
  visitStructural(Structural ast) {}
  visitConditional(Conditional ast) {}
  visitFormatter(Formatter ast) {}
  visitFunctionCall(FunctionCall ast) {}
  visitImplicitReceiver(ImplicitReceiver ast) {}
  visitKeyedAccess(KeyedAccess ast) {}
  visitLiteralArray(LiteralArray ast) {}
  visitLiteralMap(LiteralMap ast) {}
  visitLiteralPrimitive(LiteralPrimitive ast) {}
  visitMethodCall(MethodCall ast) {}
  visitPrefixNot(PrefixNot ast) {}
}
var _evalListCache = [
  [],
  [0],
  [0, 0],
  [0, 0, 0],
  [0, 0, 0, 0],
  [0, 0, 0, 0, 0]
];
evalList(context, List exps) {
  var length = exps.length;
  var result = _evalListCache[length];
  for (var i = 0; i < length; i++) {
    result[i] = exps[i].eval(context);
  }
  return result;
}
