library angular2.src.change_detection.proto_change_detector_dart;

import 'package:angular2/src/facade/lang.dart'
    show isPresent, isBlank, BaseException, Type, isString;
import 'package:angular2/src/facade/collection.dart'
    show List, ListWrapper, MapWrapper, StringMapWrapper;
import './parser/ast.dart'
    show
        AccessMember,
        Assignment,
        AST,
        ASTWithSource,
        AstVisitor,
        Binary,
        Chain,
        Structural,
        Conditional,
        Formatter,
        FunctionCall,
        ImplicitReceiver,
        Interpolation,
        KeyedAccess,
        LiteralArray,
        LiteralMap,
        LiteralPrimitive,
        MethodCall,
        PrefixNot;
import './parser/context_with_variable_bindings.dart'
    show ContextWithVariableBindings;
import './interfaces.dart' show ChangeRecord, ChangeDispatcher, ChangeDetector;
import './change_detection_util.dart' show ChangeDetectionUtil;
import './dynamic_change_detector.dart' show DynamicChangeDetector;
import './change_detection_jit_generator.dart' show ChangeDetectorJITGenerator;
import './array_changes.dart' show ArrayChanges;
import './keyvalue_changes.dart' show KeyValueChanges;
import './coalesce.dart' show coalesce;

const RECORD_TYPE_SELF = 0;
const RECORD_TYPE_CONST = 1;
const RECORD_TYPE_PRIMITIVE_OP = 2;
const RECORD_TYPE_PROPERTY = 3;
const RECORD_TYPE_INVOKE_METHOD = 4;
const RECORD_TYPE_INVOKE_CLOSURE = 5;
const RECORD_TYPE_KEYED_ACCESS = 6;
const RECORD_TYPE_INVOKE_FORMATTER = 7;
const RECORD_TYPE_STRUCTURAL_CHECK = 8;
const RECORD_TYPE_INTERPOLATE = 9;
class ProtoRecord {
  num mode;
  String name;
  dynamic funcOrValue;
  List args;
  List fixedArgs;
  num contextIndex;
  num selfIndex;
  dynamic bindingMemento;
  dynamic directiveMemento;
  bool lastInBinding;
  bool lastInDirective;
  String expressionAsString;
  ProtoRecord(num mode, String name, funcOrValue, List args, List fixedArgs,
      num contextIndex, num selfIndex, dynamic bindingMemento,
      dynamic directiveMemento, String expressionAsString, bool lastInBinding,
      bool lastInDirective) {
    this.mode = mode;
    this.name = name;
    this.funcOrValue = funcOrValue;
    this.args = args;
    this.fixedArgs = fixedArgs;
    this.contextIndex = contextIndex;
    this.selfIndex = selfIndex;
    this.bindingMemento = bindingMemento;
    this.directiveMemento = directiveMemento;
    this.lastInBinding = lastInBinding;
    this.lastInDirective = lastInDirective;
    this.expressionAsString = expressionAsString;
  }
  bool isPureFunction() {
    return identical(this.mode, RECORD_TYPE_INTERPOLATE) ||
        identical(this.mode, RECORD_TYPE_INVOKE_FORMATTER) ||
        identical(this.mode, RECORD_TYPE_PRIMITIVE_OP);
  }
}
class ProtoChangeDetector {
  addAst(AST ast, dynamic bindingMemento,
      [dynamic directiveMemento = null, bool structural = false]) {}
  ChangeDetector instantiate(dynamic dispatcher, Map formatters) {
    return null;
  }
}
class DynamicProtoChangeDetector extends ProtoChangeDetector {
  List<ProtoRecord> _records;
  ProtoRecordBuilder _recordBuilder;
  DynamicProtoChangeDetector() : super() {
    this._records = null;
    this._recordBuilder = new ProtoRecordBuilder();
  }
  addAst(AST ast, dynamic bindingMemento,
      [dynamic directiveMemento = null, bool structural = false]) {
    this._recordBuilder.addAst(
        ast, bindingMemento, directiveMemento, structural);
  }
  instantiate(dynamic dispatcher, Map formatters) {
    this._createRecordsIfNecessary();
    return new DynamicChangeDetector(dispatcher, formatters, this._records);
  }
  _createRecordsIfNecessary() {
    if (isBlank(this._records)) {
      var records = this._recordBuilder.records;
      this._records = coalesce(records);
    }
  }
}
num _jitProtoChangeDetectorClassCounter = 0;
class JitProtoChangeDetector extends ProtoChangeDetector {
  Function _factory;
  ProtoRecordBuilder _recordBuilder;
  JitProtoChangeDetector() : super() {
    this._factory = null;
    this._recordBuilder = new ProtoRecordBuilder();
  }
  addAst(AST ast, dynamic bindingMemento,
      [dynamic directiveMemento = null, bool structural = false]) {
    this._recordBuilder.addAst(
        ast, bindingMemento, directiveMemento, structural);
  }
  instantiate(dynamic dispatcher, Map formatters) {
    this._createFactoryIfNecessary();
    return this._factory(dispatcher, formatters);
  }
  _createFactoryIfNecessary() {
    if (isBlank(this._factory)) {
      var c = _jitProtoChangeDetectorClassCounter++;
      var records = coalesce(this._recordBuilder.records);
      var typeName = '''ChangeDetector${c}''';
      this._factory =
          new ChangeDetectorJITGenerator(typeName, records).generate();
    }
  }
}
class ProtoRecordBuilder {
  List<ProtoRecord> records;
  ProtoRecordBuilder() {
    this.records = [];
  }
  addAst(AST ast, dynamic bindingMemento,
      [dynamic directiveMemento = null, bool structural = false]) {
    if (structural) ast = new Structural(ast);
    var last = ListWrapper.last(this.records);
    if (isPresent(last) && last.directiveMemento == directiveMemento) {
      last.lastInDirective = false;
    }
    var pr = _ConvertAstIntoProtoRecords.convert(
        ast, bindingMemento, directiveMemento, this.records.length);
    if (!ListWrapper.isEmpty(pr)) {
      var last = ListWrapper.last(pr);
      last.lastInBinding = true;
      last.lastInDirective = true;
      this.records = ListWrapper.concat(this.records, pr);
    }
  }
}
class _ConvertAstIntoProtoRecords {
  List protoRecords;
  dynamic bindingMemento;
  dynamic directiveMemento;
  num contextIndex;
  String expressionAsString;
  _ConvertAstIntoProtoRecords(dynamic bindingMemento, dynamic directiveMemento,
      num contextIndex, String expressionAsString) {
    this.protoRecords = [];
    this.bindingMemento = bindingMemento;
    this.directiveMemento = directiveMemento;
    this.contextIndex = contextIndex;
    this.expressionAsString = expressionAsString;
  }
  static convert(AST ast, dynamic bindingMemento, dynamic directiveMemento,
      num contextIndex) {
    var c = new _ConvertAstIntoProtoRecords(
        bindingMemento, directiveMemento, contextIndex, ast.toString());
    ast.visit(c);
    return c.protoRecords;
  }
  visitImplicitReceiver(ImplicitReceiver ast) {
    return 0;
  }
  visitInterpolation(Interpolation ast) {
    var args = this._visitAll(ast.expressions);
    return this._addRecord(RECORD_TYPE_INTERPOLATE, "interpolate",
        _interpolationFn(ast.strings), args, ast.strings, 0);
  }
  visitLiteralPrimitive(LiteralPrimitive ast) {
    return this._addRecord(
        RECORD_TYPE_CONST, "literal", ast.value, [], null, 0);
  }
  visitAccessMember(AccessMember ast) {
    var receiver = ast.receiver.visit(this);
    return this._addRecord(
        RECORD_TYPE_PROPERTY, ast.name, ast.getter, [], null, receiver);
  }
  visitFormatter(Formatter ast) {
    return this._addRecord(RECORD_TYPE_INVOKE_FORMATTER, ast.name, ast.name,
        this._visitAll(ast.allArgs), null, 0);
  }
  visitMethodCall(MethodCall ast) {
    var receiver = ast.receiver.visit(this);
    var args = this._visitAll(ast.args);
    return this._addRecord(
        RECORD_TYPE_INVOKE_METHOD, ast.name, ast.fn, args, null, receiver);
  }
  visitFunctionCall(FunctionCall ast) {
    var target = ast.target.visit(this);
    var args = this._visitAll(ast.args);
    return this._addRecord(
        RECORD_TYPE_INVOKE_CLOSURE, "closure", null, args, null, target);
  }
  visitLiteralArray(LiteralArray ast) {
    var primitiveName = '''arrayFn${ast.expressions.length}''';
    return this._addRecord(RECORD_TYPE_PRIMITIVE_OP, primitiveName,
        _arrayFn(ast.expressions.length), this._visitAll(ast.expressions), null,
        0);
  }
  visitLiteralMap(LiteralMap ast) {
    return this._addRecord(RECORD_TYPE_PRIMITIVE_OP,
        _mapPrimitiveName(ast.keys), ChangeDetectionUtil.mapFn(ast.keys),
        this._visitAll(ast.values), null, 0);
  }
  visitBinary(Binary ast) {
    var left = ast.left.visit(this);
    var right = ast.right.visit(this);
    return this._addRecord(RECORD_TYPE_PRIMITIVE_OP,
        _operationToPrimitiveName(ast.operation),
        _operationToFunction(ast.operation), [left, right], null, 0);
  }
  visitPrefixNot(PrefixNot ast) {
    var exp = ast.expression.visit(this);
    return this._addRecord(RECORD_TYPE_PRIMITIVE_OP, "operation_negate",
        ChangeDetectionUtil.operation_negate, [exp], null, 0);
  }
  visitConditional(Conditional ast) {
    var c = ast.condition.visit(this);
    var t = ast.trueExp.visit(this);
    var f = ast.falseExp.visit(this);
    return this._addRecord(RECORD_TYPE_PRIMITIVE_OP, "cond",
        ChangeDetectionUtil.cond, [c, t, f], null, 0);
  }
  visitStructural(Structural ast) {
    var value = ast.value.visit(this);
    return this._addRecord(
        RECORD_TYPE_STRUCTURAL_CHECK, "structural", null, [], null, value);
  }
  visitKeyedAccess(KeyedAccess ast) {
    var obj = ast.obj.visit(this);
    var key = ast.key.visit(this);
    return this._addRecord(RECORD_TYPE_KEYED_ACCESS, "keyedAccess",
        ChangeDetectionUtil.keyedAccess, [key], null, obj);
  }
  _visitAll(List asts) {
    var res = ListWrapper.createFixedSize(asts.length);
    for (var i = 0; i < asts.length; ++i) {
      res[i] = asts[i].visit(this);
    }
    return res;
  }
  _addRecord(type, name, funcOrValue, args, fixedArgs, context) {
    var selfIndex = ++this.contextIndex;
    ListWrapper.push(this.protoRecords, new ProtoRecord(type, name, funcOrValue,
        args, fixedArgs, context, selfIndex, this.bindingMemento,
        this.directiveMemento, this.expressionAsString, false, false));
    return selfIndex;
  }
}
Function _arrayFn(num length) {
  switch (length) {
    case 0:
      return ChangeDetectionUtil.arrayFn0;
    case 1:
      return ChangeDetectionUtil.arrayFn1;
    case 2:
      return ChangeDetectionUtil.arrayFn2;
    case 3:
      return ChangeDetectionUtil.arrayFn3;
    case 4:
      return ChangeDetectionUtil.arrayFn4;
    case 5:
      return ChangeDetectionUtil.arrayFn5;
    case 6:
      return ChangeDetectionUtil.arrayFn6;
    case 7:
      return ChangeDetectionUtil.arrayFn7;
    case 8:
      return ChangeDetectionUtil.arrayFn8;
    case 9:
      return ChangeDetectionUtil.arrayFn9;
    default:
      throw new BaseException(
          '''Does not support literal maps with more than 9 elements''');
  }
}
_mapPrimitiveName(List keys) {
  var stringifiedKeys = ListWrapper.join(ListWrapper.map(keys, (k) {
    return isString(k) ? '''\"${k}\"''' : '''${k}''';
  }), ", ");
  return '''mapFn([${stringifiedKeys}])''';
}
String _operationToPrimitiveName(String operation) {
  switch (operation) {
    case '+':
      return "operation_add";
    case '-':
      return "operation_subtract";
    case '*':
      return "operation_multiply";
    case '/':
      return "operation_divide";
    case '%':
      return "operation_remainder";
    case '==':
      return "operation_equals";
    case '!=':
      return "operation_not_equals";
    case '<':
      return "operation_less_then";
    case '>':
      return "operation_greater_then";
    case '<=':
      return "operation_less_or_equals_then";
    case '>=':
      return "operation_greater_or_equals_then";
    case '&&':
      return "operation_logical_and";
    case '||':
      return "operation_logical_or";
    default:
      throw new BaseException('''Unsupported operation ${operation}''');
  }
}
Function _operationToFunction(String operation) {
  switch (operation) {
    case '+':
      return ChangeDetectionUtil.operation_add;
    case '-':
      return ChangeDetectionUtil.operation_subtract;
    case '*':
      return ChangeDetectionUtil.operation_multiply;
    case '/':
      return ChangeDetectionUtil.operation_divide;
    case '%':
      return ChangeDetectionUtil.operation_remainder;
    case '==':
      return ChangeDetectionUtil.operation_equals;
    case '!=':
      return ChangeDetectionUtil.operation_not_equals;
    case '<':
      return ChangeDetectionUtil.operation_less_then;
    case '>':
      return ChangeDetectionUtil.operation_greater_then;
    case '<=':
      return ChangeDetectionUtil.operation_less_or_equals_then;
    case '>=':
      return ChangeDetectionUtil.operation_greater_or_equals_then;
    case '&&':
      return ChangeDetectionUtil.operation_logical_and;
    case '||':
      return ChangeDetectionUtil.operation_logical_or;
    default:
      throw new BaseException('''Unsupported operation ${operation}''');
  }
}
s(v) {
  return isPresent(v) ? '''${v}''' : '';
}
_interpolationFn(List strings) {
  var length = strings.length;
  var c0 = length > 0 ? strings[0] : null;
  var c1 = length > 1 ? strings[1] : null;
  var c2 = length > 2 ? strings[2] : null;
  var c3 = length > 3 ? strings[3] : null;
  var c4 = length > 4 ? strings[4] : null;
  var c5 = length > 5 ? strings[5] : null;
  var c6 = length > 6 ? strings[6] : null;
  var c7 = length > 7 ? strings[7] : null;
  var c8 = length > 8 ? strings[8] : null;
  var c9 = length > 9 ? strings[9] : null;
  switch (length - 1) {
    case 1:
      return (a1) {
        return c0 + s(a1) + c1;
      };
    case 2:
      return (a1, a2) {
        return c0 + s(a1) + c1 + s(a2) + c2;
      };
    case 3:
      return (a1, a2, a3) {
        return c0 + s(a1) + c1 + s(a2) + c2 + s(a3) + c3;
      };
    case 4:
      return (a1, a2, a3, a4) {
        return c0 + s(a1) + c1 + s(a2) + c2 + s(a3) + c3 + s(a4) + c4;
      };
    case 5:
      return (a1, a2, a3, a4, a5) {
        return c0 +
            s(a1) +
            c1 +
            s(a2) +
            c2 +
            s(a3) +
            c3 +
            s(a4) +
            c4 +
            s(a5) +
            c5;
      };
    case 6:
      return (a1, a2, a3, a4, a5, a6) {
        return c0 +
            s(a1) +
            c1 +
            s(a2) +
            c2 +
            s(a3) +
            c3 +
            s(a4) +
            c4 +
            s(a5) +
            c5 +
            s(a6) +
            c6;
      };
    case 7:
      return (a1, a2, a3, a4, a5, a6, a7) {
        return c0 +
            s(a1) +
            c1 +
            s(a2) +
            c2 +
            s(a3) +
            c3 +
            s(a4) +
            c4 +
            s(a5) +
            c5 +
            s(a6) +
            c6 +
            s(a7) +
            c7;
      };
    case 8:
      return (a1, a2, a3, a4, a5, a6, a7, a8) {
        return c0 +
            s(a1) +
            c1 +
            s(a2) +
            c2 +
            s(a3) +
            c3 +
            s(a4) +
            c4 +
            s(a5) +
            c5 +
            s(a6) +
            c6 +
            s(a7) +
            c7 +
            s(a8) +
            c8;
      };
    case 9:
      return (a1, a2, a3, a4, a5, a6, a7, a8, a9) {
        return c0 +
            s(a1) +
            c1 +
            s(a2) +
            c2 +
            s(a3) +
            c3 +
            s(a4) +
            c4 +
            s(a5) +
            c5 +
            s(a6) +
            c6 +
            s(a7) +
            c7 +
            s(a8) +
            c8 +
            s(a9) +
            c9;
      };
    default:
      throw new BaseException('''Does not support more than 9 expressions''');
  }
}
