library angular2.src.core.compiler.compiler_dart;

import 'package:angular2/src/facade/lang.dart'
    show Type, isBlank, isPresent, BaseException, normalizeBlank, stringify;
import 'package:angular2/src/facade/async.dart' show Future, PromiseWrapper;
import 'package:angular2/src/facade/collection.dart'
    show List, ListWrapper, Map, MapWrapper;
import 'package:angular2/src/facade/dom.dart' show DOM, Element;
import 'package:angular2/change_detection.dart' show ChangeDetection, Parser;
import './directive_metadata_reader.dart' show DirectiveMetadataReader;
import './view.dart' show ProtoView;
import './pipeline/compile_pipeline.dart' show CompilePipeline;
import './pipeline/compile_element.dart' show CompileElement;
import './pipeline/default_steps.dart' show createDefaultSteps;
import './template_loader.dart' show TemplateLoader;
import './template_resolver.dart' show TemplateResolver;
import './directive_metadata.dart' show DirectiveMetadata;
import '../annotations/annotations.dart' show Component;
import '../annotations/template.dart' show Template;
import './shadow_dom_emulation/content_tag.dart' show Content;
import './shadow_dom_strategy.dart' show ShadowDomStrategy;
import './pipeline/compile_step.dart' show CompileStep;

class CompilerCache {
  Map _cache;
  CompilerCache() {
    this._cache = MapWrapper.create();
  }
  set(Type component, ProtoView protoView) {
    MapWrapper.set(this._cache, component, protoView);
  }
  ProtoView get(Type component) {
    var result = MapWrapper.get(this._cache, component);
    return normalizeBlank(result);
  }
  clear() {
    MapWrapper.clear(this._cache);
  }
}
class Compiler {
  DirectiveMetadataReader _reader;
  Parser _parser;
  CompilerCache _compilerCache;
  ChangeDetection _changeDetection;
  TemplateLoader _templateLoader;
  Map<Type, Future> _compiling;
  ShadowDomStrategy _shadowDomStrategy;
  List<DirectiveMetadata> _shadowDomDirectives;
  TemplateResolver _templateResolver;
  Compiler(ChangeDetection changeDetection, TemplateLoader templateLoader,
      DirectiveMetadataReader reader, Parser parser, CompilerCache cache,
      ShadowDomStrategy shadowDomStrategy, TemplateResolver templateResolver) {
    this._changeDetection = changeDetection;
    this._reader = reader;
    this._parser = parser;
    this._compilerCache = cache;
    this._templateLoader = templateLoader;
    this._compiling = MapWrapper.create();
    this._shadowDomStrategy = shadowDomStrategy;
    this._shadowDomDirectives = [];
    var types = shadowDomStrategy.polyfillDirectives();
    for (var i = 0; i < types.length; i++) {
      ListWrapper.push(this._shadowDomDirectives, reader.read(types[i]));
    }
    this._templateResolver = templateResolver;
  }
  List<CompileStep> createSteps(Type component, Template template) {
    var dirMetadata = [];
    var tplMetadata = ListWrapper.map(this._flattenDirectives(template), (d) {
      return this._reader.read(d);
    });
    dirMetadata = ListWrapper.concat(dirMetadata, tplMetadata);
    dirMetadata = ListWrapper.concat(dirMetadata, this._shadowDomDirectives);
    var cmpMetadata = this._reader.read(component);
    return createDefaultSteps(this._changeDetection, this._parser, cmpMetadata,
        dirMetadata, this._shadowDomStrategy);
  }
  Future<ProtoView> compile(Type component) {
    var protoView = this._compile(component);
    return PromiseWrapper.isPromise(protoView)
        ? protoView
        : PromiseWrapper.resolve(protoView);
  }
  _compile(Type component) {
    var protoView = this._compilerCache.get(component);
    if (isPresent(protoView)) {
      return protoView;
    }
    var pvPromise = MapWrapper.get(this._compiling, component);
    if (isPresent(pvPromise)) {
      return pvPromise;
    }
    var template = this._templateResolver.resolve(component);
    var tplElement = this._templateLoader.load(template);
    if (PromiseWrapper.isPromise(tplElement)) {
      pvPromise = PromiseWrapper.then(tplElement, (el) {
        return this._compileTemplate(template, el, component);
      }, (_) {
        throw new BaseException(
            '''Failed to load the template for ${stringify(component)}''');
      });
      MapWrapper.set(this._compiling, component, pvPromise);
      return pvPromise;
    }
    return this._compileTemplate(template, tplElement, component);
  }
  _compileTemplate(Template template, Element tplElement, Type component) {
    var pipeline = new CompilePipeline(this.createSteps(component, template));
    var compileElements = pipeline.process(tplElement);
    var protoView = compileElements[0].inheritedProtoView;
    this._compilerCache.set(component, protoView);
    MapWrapper.delete(this._compiling, component);
    var nestedPVPromises = [];
    for (var i = 0; i < compileElements.length; i++) {
      var ce = compileElements[i];
      if (isPresent(ce.componentDirective)) {
        this._compileNestedProtoView(ce, nestedPVPromises);
      }
    }
    if (nestedPVPromises.length > 0) {
      return PromiseWrapper.then(PromiseWrapper.all(nestedPVPromises), (_) {
        return protoView;
      }, (e) {
        throw new BaseException(
            '''${e.message} -> Failed to compile ${stringify(component)}''');
      });
    }
    return protoView;
  }
  _compileNestedProtoView(CompileElement ce, List<Future> promises) {
    var protoView = this._compile(ce.componentDirective.type);
    if (PromiseWrapper.isPromise(protoView)) {
      ListWrapper.push(promises, protoView);
      protoView.then((protoView) {
        ce.inheritedElementBinder.nestedProtoView = protoView;
      });
    } else {
      ce.inheritedElementBinder.nestedProtoView = protoView;
    }
  }
  List<Type> _flattenDirectives(Template template) {
    if (isBlank(template.directives)) return [];
    var directives = [];
    this._flattenList(template.directives, directives);
    return directives;
  }
  _flattenList(List<dynamic> tree, List<Type> out) {
    for (var i = 0; i < tree.length; i++) {
      var item = tree[i];
      if (ListWrapper.isList(item)) {
        this._flattenList(item, out);
      } else {
        ListWrapper.push(out, item);
      }
    }
  }
}
