library angular2.src.core.compiler.pipeline.element_binder_builder_dart;

import 'package:angular2/src/facade/lang.dart'
    show
        int,
        isPresent,
        isBlank,
        Type,
        BaseException,
        StringWrapper,
        RegExpWrapper,
        isString,
        stringify;
import 'package:angular2/src/facade/dom.dart' show Element, DOM;
import 'package:angular2/src/facade/collection.dart'
    show ListWrapper, List, MapWrapper, StringMapWrapper;
import 'package:angular2/src/reflection/reflection.dart' show reflector;
import 'package:angular2/change_detection.dart'
    show Parser, ProtoChangeDetector;
import '../../annotations/annotations.dart' show Component, Directive;
import '../directive_metadata.dart' show DirectiveMetadata;
import '../view.dart'
    show ProtoView, ElementPropertyMemento, DirectivePropertyMemento;
import '../element_injector.dart' show ProtoElementInjector;
import '../element_binder.dart' show ElementBinder;
import './compile_step.dart' show CompileStep;
import './compile_element.dart' show CompileElement;
import './compile_control.dart' show CompileControl;

var DOT_REGEXP = RegExpWrapper.create('\\.');
const ARIA_PREFIX = 'aria-';
var ariaSettersCache = StringMapWrapper.create();
ariaSetterFactory(String attrName) {
  var setterFn = StringMapWrapper.get(ariaSettersCache, attrName);
  if (isBlank(setterFn)) {
    setterFn = (Element element, value) {
      if (isPresent(value)) {
        DOM.setAttribute(element, attrName, stringify(value));
      } else {
        DOM.removeAttribute(element, attrName);
      }
    };
    StringMapWrapper.set(ariaSettersCache, attrName, setterFn);
  }
  return setterFn;
}
const CLASS_PREFIX = 'class.';
var classSettersCache = StringMapWrapper.create();
classSetterFactory(String className) {
  var setterFn = StringMapWrapper.get(classSettersCache, className);
  if (isBlank(setterFn)) {
    setterFn = (Element element, value) {
      if (value) {
        DOM.addClass(element, className);
      } else {
        DOM.removeClass(element, className);
      }
    };
    StringMapWrapper.set(classSettersCache, className, setterFn);
  }
  return setterFn;
}
const STYLE_PREFIX = 'style.';
var styleSettersCache = StringMapWrapper.create();
styleSetterFactory(String styleName, String stylesuffix) {
  var cacheKey = styleName + stylesuffix;
  var setterFn = StringMapWrapper.get(styleSettersCache, cacheKey);
  if (isBlank(setterFn)) {
    setterFn = (Element element, value) {
      var valAsStr;
      if (isPresent(value)) {
        valAsStr = stringify(value);
        DOM.setStyle(element, styleName, valAsStr + stylesuffix);
      } else {
        DOM.removeStyle(element, styleName);
      }
    };
    StringMapWrapper.set(classSettersCache, cacheKey, setterFn);
  }
  return setterFn;
}
const ROLE_ATTR = 'role';
roleSetter(Element element, value) {
  if (isString(value)) {
    DOM.setAttribute(element, ROLE_ATTR, value);
  } else {
    DOM.removeAttribute(element, ROLE_ATTR);
    if (isPresent(value)) {
      throw new BaseException(
          "Invalid role attribute, only string values are allowed, got '" +
              stringify(value) +
              "'");
    }
  }
}
class ElementBinderBuilder extends CompileStep {
  Parser _parser;
  dynamic _compilationUnit;
  ElementBinderBuilder(Parser parser, dynamic compilationUnit) : super() {
    this._parser = parser;
    this._compilationUnit = compilationUnit;
  }
  process(
      CompileElement parent, CompileElement current, CompileControl control) {
    var elementBinder = null;
    if (current.hasBindings) {
      var protoView = current.inheritedProtoView;
      var protoInjectorWasBuilt = isBlank(parent)
          ? true
          : !identical(current.inheritedProtoElementInjector,
              parent.inheritedProtoElementInjector);
      var currentProtoElementInjector =
          protoInjectorWasBuilt ? current.inheritedProtoElementInjector : null;
      elementBinder = protoView.bindElement(currentProtoElementInjector,
          current.componentDirective, current.viewportDirective);
      if (isPresent(current.textNodeBindings)) {
        this._bindTextNodes(protoView, current);
      }
      if (isPresent(current.propertyBindings)) {
        this._bindElementProperties(protoView, current);
      }
      if (isPresent(current.eventBindings)) {
        this._bindEvents(protoView, current);
      }
      this._bindDirectiveProperties(current.getAllDirectives(), current);
    } else if (isPresent(parent)) {
      elementBinder = parent.inheritedElementBinder;
    }
    current.inheritedElementBinder = elementBinder;
  }
  _bindTextNodes(protoView, compileElement) {
    MapWrapper.forEach(compileElement.textNodeBindings,
        (expression, indexInParent) {
      protoView.bindTextNode(indexInParent, expression);
    });
  }
  _bindElementProperties(protoView, compileElement) {
    MapWrapper.forEach(compileElement.propertyBindings, (expression, property) {
      var setterFn;
      var styleParts;
      var styleSuffix;
      if (StringWrapper.startsWith(property, ARIA_PREFIX)) {
        setterFn = ariaSetterFactory(property);
      } else if (StringWrapper.equals(property, ROLE_ATTR)) {
        setterFn = roleSetter;
      } else if (StringWrapper.startsWith(property, CLASS_PREFIX)) {
        setterFn = classSetterFactory(
            StringWrapper.substring(property, CLASS_PREFIX.length));
      } else if (StringWrapper.startsWith(property, STYLE_PREFIX)) {
        styleParts = StringWrapper.split(property, DOT_REGEXP);
        styleSuffix =
            styleParts.length > 2 ? ListWrapper.get(styleParts, 2) : '';
        setterFn =
            styleSetterFactory(ListWrapper.get(styleParts, 1), styleSuffix);
      } else if (DOM.hasProperty(compileElement.element, property)) {
        setterFn = reflector.setter(property);
      }
      if (isPresent(setterFn)) {
        protoView.bindElementProperty(expression.ast, property, setterFn);
      }
    });
  }
  _bindEvents(protoView, compileElement) {
    MapWrapper.forEach(compileElement.eventBindings, (expression, eventName) {
      protoView.bindEvent(eventName, expression);
    });
  }
  _bindDirectiveProperties(
      List<DirectiveMetadata> directives, CompileElement compileElement) {
    var protoView = compileElement.inheritedProtoView;
    for (var directiveIndex = 0;
        directiveIndex < directives.length;
        directiveIndex++) {
      var directive = ListWrapper.get(directives, directiveIndex);
      var annotation = directive.annotation;
      if (isBlank(annotation.bind)) continue;
      var _this = this;
      StringMapWrapper.forEach(annotation.bind, (dirProp, elProp) {
        var expression = isPresent(compileElement.propertyBindings)
            ? MapWrapper.get(compileElement.propertyBindings, elProp)
            : null;
        if (isBlank(expression)) {
          var attributeValue = MapWrapper.get(compileElement.attrs(), elProp);
          if (isPresent(attributeValue)) {
            expression = _this._parser.wrapLiteralPrimitive(
                attributeValue, _this._compilationUnit);
          } else {
            throw new BaseException("No element binding found for property '" +
                elProp +
                "' which is required by directive '" +
                stringify(directive.type) +
                "'");
          }
        }
        var len = dirProp.length;
        var dirBindingName = dirProp;
        var isContentWatch = identical(dirProp[len - 2], '[') &&
            identical(dirProp[len - 1], ']');
        if (isContentWatch) dirBindingName = dirProp.substring(0, len - 2);
        protoView.bindDirectiveProperty(directiveIndex, expression,
            dirBindingName, reflector.setter(dirBindingName), isContentWatch);
      });
    }
  }
}
