library angular2.src.core.compiler.selector_dart;

import 'package:angular2/src/facade/collection.dart'
    show List, Map, ListWrapper, MapWrapper;
import 'package:angular2/src/facade/lang.dart'
    show isPresent, isBlank, RegExpWrapper, RegExpMatcherWrapper, StringWrapper;

const _EMPTY_ATTR_VALUE = '';
var _SELECTOR_REGEXP = RegExpWrapper.create('^([-\\w]+)|' +
    '(?:\\.([-\\w]+))|' +
    '(?:\\[([-\\w*]+)(?:=([^\\]]*))?\\])');
class CssSelector {
  String element;
  List classNames;
  List attrs;
  static CssSelector parse(String selector) {
    var cssSelector = new CssSelector();
    var matcher = RegExpWrapper.matcher(_SELECTOR_REGEXP, selector);
    var match;
    while (isPresent(match = RegExpMatcherWrapper.next(matcher))) {
      if (isPresent(match[1])) {
        cssSelector.setElement(match[1]);
      }
      if (isPresent(match[2])) {
        cssSelector.addClassName(match[2]);
      }
      if (isPresent(match[3])) {
        cssSelector.addAttribute(match[3], match[4]);
      }
    }
    return cssSelector;
  }
  CssSelector() {
    this.element = null;
    this.classNames = ListWrapper.create();
    this.attrs = ListWrapper.create();
  }
  setElement([String element = null]) {
    if (isPresent(element)) {
      element = element.toLowerCase();
    }
    this.element = element;
  }
  addAttribute(String name, [String value = _EMPTY_ATTR_VALUE]) {
    ListWrapper.push(this.attrs, name.toLowerCase());
    if (isPresent(value)) {
      value = value.toLowerCase();
    } else {
      value = _EMPTY_ATTR_VALUE;
    }
    ListWrapper.push(this.attrs, value);
  }
  addClassName(String name) {
    ListWrapper.push(this.classNames, name.toLowerCase());
  }
  String toString() {
    var res = '';
    if (isPresent(this.element)) {
      res += this.element;
    }
    if (isPresent(this.classNames)) {
      for (var i = 0; i < this.classNames.length; i++) {
        res += '.' + this.classNames[i];
      }
    }
    if (isPresent(this.attrs)) {
      for (var i = 0; i < this.attrs.length;) {
        var attrName = this.attrs[i++];
        var attrValue = this.attrs[i++];
        res += '[' + attrName;
        if (attrValue.length > 0) {
          res += '=' + attrValue;
        }
        res += ']';
      }
    }
    return res;
  }
}
class SelectorMatcher {
  Map _elementMap;
  Map _elementPartialMap;
  Map _classMap;
  Map _classPartialMap;
  Map _attrValueMap;
  Map _attrValuePartialMap;
  SelectorMatcher() {
    this._elementMap = MapWrapper.create();
    this._elementPartialMap = MapWrapper.create();
    this._classMap = MapWrapper.create();
    this._classPartialMap = MapWrapper.create();
    this._attrValueMap = MapWrapper.create();
    this._attrValuePartialMap = MapWrapper.create();
  }
  addSelectable(CssSelector cssSelector, selectable) {
    var matcher = this;
    var element = cssSelector.element;
    var classNames = cssSelector.classNames;
    var attrs = cssSelector.attrs;
    if (isPresent(element)) {
      var isTerminal =
          identical(attrs.length, 0) && identical(classNames.length, 0);
      if (isTerminal) {
        this._addTerminal(matcher._elementMap, element, selectable);
      } else {
        matcher = this._addPartial(matcher._elementPartialMap, element);
      }
    }
    if (isPresent(classNames)) {
      for (var index = 0; index < classNames.length; index++) {
        var isTerminal = identical(attrs.length, 0) &&
            identical(index, classNames.length - 1);
        var className = classNames[index];
        if (isTerminal) {
          this._addTerminal(matcher._classMap, className, selectable);
        } else {
          matcher = this._addPartial(matcher._classPartialMap, className);
        }
      }
    }
    if (isPresent(attrs)) {
      for (var index = 0; index < attrs.length;) {
        var isTerminal = identical(index, attrs.length - 2);
        var attrName = attrs[index++];
        var attrValue = attrs[index++];
        var map =
            isTerminal ? matcher._attrValueMap : matcher._attrValuePartialMap;
        var valuesMap = MapWrapper.get(map, attrName);
        if (isBlank(valuesMap)) {
          valuesMap = MapWrapper.create();
          MapWrapper.set(map, attrName, valuesMap);
        }
        if (isTerminal) {
          this._addTerminal(valuesMap, attrValue, selectable);
        } else {
          matcher = this._addPartial(valuesMap, attrValue);
        }
      }
    }
  }
  _addTerminal(Map<String, String> map, String name, selectable) {
    var terminalList = MapWrapper.get(map, name);
    if (isBlank(terminalList)) {
      terminalList = ListWrapper.create();
      MapWrapper.set(map, name, terminalList);
    }
    ListWrapper.push(terminalList, selectable);
  }
  _addPartial(Map<String, String> map, String name) {
    var matcher = MapWrapper.get(map, name);
    if (isBlank(matcher)) {
      matcher = new SelectorMatcher();
      MapWrapper.set(map, name, matcher);
    }
    return matcher;
  }
  match(CssSelector cssSelector, Function matchedCallback) {
    var element = cssSelector.element;
    var classNames = cssSelector.classNames;
    var attrs = cssSelector.attrs;
    this._matchTerminal(this._elementMap, element, matchedCallback);
    this._matchPartial(
        this._elementPartialMap, element, cssSelector, matchedCallback);
    if (isPresent(classNames)) {
      for (var index = 0; index < classNames.length; index++) {
        var className = classNames[index];
        this._matchTerminal(this._classMap, className, matchedCallback);
        this._matchPartial(
            this._classPartialMap, className, cssSelector, matchedCallback);
      }
    }
    if (isPresent(attrs)) {
      for (var index = 0; index < attrs.length;) {
        var attrName = attrs[index++];
        var attrValue = attrs[index++];
        var valuesMap = MapWrapper.get(this._attrValueMap, attrName);
        if (!StringWrapper.equals(attrValue, _EMPTY_ATTR_VALUE)) {
          this._matchTerminal(valuesMap, _EMPTY_ATTR_VALUE, matchedCallback);
        }
        this._matchTerminal(valuesMap, attrValue, matchedCallback);
        valuesMap = MapWrapper.get(this._attrValuePartialMap, attrName);
        this._matchPartial(valuesMap, attrValue, cssSelector, matchedCallback);
      }
    }
  }
  _matchTerminal([Map<String, String> map = null, name, matchedCallback]) {
    if (isBlank(map) || isBlank(name)) {
      return;
    }
    var selectables = MapWrapper.get(map, name);
    if (isBlank(selectables)) {
      return;
    }
    for (var index = 0; index < selectables.length; index++) {
      matchedCallback(selectables[index]);
    }
  }
  _matchPartial(
      [Map<String, String> map = null, name, cssSelector, matchedCallback]) {
    if (isBlank(map) || isBlank(name)) {
      return;
    }
    var nestedSelector = MapWrapper.get(map, name);
    if (isBlank(nestedSelector)) {
      return;
    }
    nestedSelector.match(cssSelector, matchedCallback);
  }
}
