library angular2.test.core.compiler.integration_spec_dart;

import 'package:angular2/test_lib.dart'
    show describe, xit, it, expect, beforeEach, ddescribe, iit, el;
import 'package:angular2/src/facade/dom.dart' show DOM;
import 'package:angular2/src/facade/collection.dart' show Map, MapWrapper;
import 'package:angular2/src/facade/lang.dart' show Type, isPresent;
import 'package:angular2/di.dart' show Injector;
import 'package:angular2/change_detection.dart'
    show Lexer, Parser, ChangeDetector, dynamicChangeDetection;
import 'package:angular2/src/core/compiler/compiler.dart'
    show Compiler, CompilerCache;
import 'package:angular2/src/core/compiler/directive_metadata_reader.dart'
    show DirectiveMetadataReader;
import 'package:angular2/src/core/compiler/shadow_dom_strategy.dart'
    show NativeShadowDomStrategy;
import 'package:angular2/src/core/compiler/template_loader.dart'
    show TemplateLoader;
import 'package:angular2/src/core/compiler/template_resolver.dart'
    show TemplateResolver;
import 'package:angular2/src/core/compiler/binding_propagation_config.dart'
    show BindingPropagationConfig;
import 'package:angular2/src/core/annotations/annotations.dart'
    show Decorator, Component, Viewport;
import 'package:angular2/src/core/annotations/template.dart' show Template;
import 'package:angular2/src/core/compiler/view_container.dart'
    show ViewContainer;

main() {
  describe('integration tests', () {
    var compiler;
    var tplResolver;
    beforeEach(() {
      tplResolver = new FakeTemplateResolver();
      compiler = new Compiler(dynamicChangeDetection, new TemplateLoader(null),
          new DirectiveMetadataReader(), new Parser(new Lexer()),
          new CompilerCache(), new NativeShadowDomStrategy(), tplResolver);
    });
    describe('react to record changes', () {
      var view;
      var ctx;
      var cd;
      createView(pv) {
        ctx = new MyComp();
        view = pv.instantiate(null, null);
        view.hydrate(new Injector([]), null, ctx);
        cd = view.changeDetector;
      }
      it('should consume text node changes', (done) {
        tplResolver.setTemplate(
            MyComp, new Template(inline: '<div>{{ctxProp}}</div>'));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          ctx.ctxProp = 'Hello World!';
          cd.detectChanges();
          expect(DOM.getInnerHTML(view.nodes[0])).toEqual('Hello World!');
          done();
        });
      });
      it('should consume element binding changes', (done) {
        tplResolver.setTemplate(
            MyComp, new Template(inline: '<div [id]="ctxProp"></div>'));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          ctx.ctxProp = 'Hello World!';
          cd.detectChanges();
          expect(view.nodes[0].id).toEqual('Hello World!');
          done();
        });
      });
      it('should consume binding to aria-* attributes', (done) {
        tplResolver.setTemplate(
            MyComp, new Template(inline: '<div [aria-label]="ctxProp"></div>'));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          ctx.ctxProp = 'Initial aria label';
          cd.detectChanges();
          expect(DOM.getAttribute(view.nodes[0], 'aria-label'))
              .toEqual('Initial aria label');
          ctx.ctxProp = 'Changed aria label';
          cd.detectChanges();
          expect(DOM.getAttribute(view.nodes[0], 'aria-label'))
              .toEqual('Changed aria label');
          done();
        });
      });
      it('should consume directive watch expression change.', (done) {
        var tpl = '<div>' +
            '<div my-dir [elprop]="ctxProp"></div>' +
            '<div my-dir elprop="Hi there!"></div>' +
            '<div my-dir elprop="Hi {{\'there!\'}}"></div>' +
            '<div my-dir elprop="One more {{ctxProp}}"></div>' +
            '</div>';
        tplResolver.setTemplate(
            MyComp, new Template(inline: tpl, directives: [MyDir]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          ctx.ctxProp = 'Hello World!';
          cd.detectChanges();
          expect(view.elementInjectors[0].get(MyDir).dirProp)
              .toEqual('Hello World!');
          expect(view.elementInjectors[1].get(MyDir).dirProp)
              .toEqual('Hi there!');
          expect(view.elementInjectors[2].get(MyDir).dirProp)
              .toEqual('Hi there!');
          expect(view.elementInjectors[3].get(MyDir).dirProp)
              .toEqual('One more Hello World!');
          done();
        });
      });
      it('should support nested components.', (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<child-cmp></child-cmp>', directives: [ChildComp]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          cd.detectChanges();
          expect(view.nodes[0].shadowRoot.childNodes[0].nodeValue)
              .toEqual('hello');
          done();
        });
      });
      it('should support different directive types on a single node', (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<child-cmp my-dir [elprop]="ctxProp"></child-cmp>',
            directives: [MyDir, ChildComp]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          ctx.ctxProp = 'Hello World!';
          cd.detectChanges();
          var elInj = view.elementInjectors[0];
          expect(elInj.get(MyDir).dirProp).toEqual('Hello World!');
          expect(elInj.get(ChildComp).dirProp).toEqual(null);
          done();
        });
      });
      it('should support template directives via `<template>` elements.',
          (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<div><template some-viewport var-greeting="some-tmpl"><copy-me>{{greeting}}</copy-me></template></div>',
            directives: [SomeViewport]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          cd.detectChanges();
          var childNodesOfWrapper = view.nodes[0].childNodes;
          expect(childNodesOfWrapper.length).toBe(3);
          expect(childNodesOfWrapper[1].childNodes[0].nodeValue)
              .toEqual('hello');
          expect(childNodesOfWrapper[2].childNodes[0].nodeValue)
              .toEqual('again');
          done();
        });
      });
      it('should support template directives via `template` attribute.',
          (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<div><copy-me template="some-viewport: var greeting=some-tmpl">{{greeting}}</copy-me></div>',
            directives: [SomeViewport]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          cd.detectChanges();
          var childNodesOfWrapper = view.nodes[0].childNodes;
          expect(childNodesOfWrapper.length).toBe(3);
          expect(childNodesOfWrapper[1].childNodes[0].nodeValue)
              .toEqual('hello');
          expect(childNodesOfWrapper[2].childNodes[0].nodeValue)
              .toEqual('again');
          done();
        });
      });
      it('should assign the component instance to a var-', (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<p><child-cmp var-alice></child-cmp></p>',
            directives: [ChildComp]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          expect(view.contextWithLocals).not.toBe(null);
          expect(view.contextWithLocals.get('alice'))
              .toBeAnInstanceOf(ChildComp);
          done();
        });
      });
      it('should assign two component instances each with a var-', (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<p><child-cmp var-alice></child-cmp><child-cmp var-bob></p>',
            directives: [ChildComp]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          expect(view.contextWithLocals).not.toBe(null);
          expect(view.contextWithLocals.get('alice'))
              .toBeAnInstanceOf(ChildComp);
          expect(view.contextWithLocals.get('bob')).toBeAnInstanceOf(ChildComp);
          expect(view.contextWithLocals.get('alice')).not
              .toBe(view.contextWithLocals.get('bob'));
          done();
        });
      });
      it('should assign the component instance to a var- with shorthand syntax',
          (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<child-cmp #alice></child-cmp>', directives: [ChildComp]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          expect(view.contextWithLocals).not.toBe(null);
          expect(view.contextWithLocals.get('alice'))
              .toBeAnInstanceOf(ChildComp);
          done();
        });
      });
      it('should assign the element instance to a user-defined variable',
          (done) {
        tplResolver.setTemplate(MyComp,
            new Template(inline: '<p><div var-alice><i>Hello</i></div></p>'));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          expect(view.contextWithLocals).not.toBe(null);
          var value = view.contextWithLocals.get('alice');
          expect(value).not.toBe(null);
          expect(value.tagName).toEqual('DIV');
          done();
        });
      });
      it('should provide binding configuration config to the component',
          (done) {
        tplResolver.setTemplate(MyComp, new Template(
            inline: '<push-cmp #cmp></push-cmp>',
            directives: [[[PushBasedComp]]]));
        compiler.compile(MyComp).then((pv) {
          createView(pv);
          var cmp = view.contextWithLocals.get('cmp');
          cd.detectChanges();
          expect(cmp.numberOfChecks).toEqual(1);
          cd.detectChanges();
          expect(cmp.numberOfChecks).toEqual(1);
          cmp.propagate();
          cd.detectChanges();
          expect(cmp.numberOfChecks).toEqual(2);
          done();
        });
      });
    });
  });
}
@Decorator(selector: '[my-dir]', bind: const {'elprop': 'dirProp'})
class MyDir {
  String dirProp;
  MyDir() {
    this.dirProp = '';
  }
}
@Component(selector: 'push-cmp')
@Template(inline: '{{field}}')
class PushBasedComp {
  num numberOfChecks;
  BindingPropagationConfig bpc;
  PushBasedComp(BindingPropagationConfig bpc) {
    this.numberOfChecks = 0;
    this.bpc = bpc;
    bpc.shouldBePropagated();
  }
  get field {
    this.numberOfChecks++;
    return "fixed";
  }
  propagate() {
    this.bpc.shouldBePropagatedFromRoot();
  }
}
@Component()
class MyComp {
  String ctxProp;
  MyComp() {
    this.ctxProp = 'initial value';
  }
}
@Component(selector: 'child-cmp', componentServices: const [MyService])
@Template(directives: const [MyDir], inline: '{{ctxProp}}')
class ChildComp {
  String ctxProp;
  String dirProp;
  ChildComp(MyService service) {
    this.ctxProp = service.greeting;
    this.dirProp = null;
  }
}
@Viewport(selector: '[some-viewport]')
class SomeViewport {
  SomeViewport(ViewContainer container) {
    container.create().setLocal('some-tmpl', 'hello');
    container.create().setLocal('some-tmpl', 'again');
  }
}
class MyService {
  String greeting;
  MyService() {
    this.greeting = 'hello';
  }
}
class FakeTemplateResolver extends TemplateResolver {
  Map _cmpTemplates;
  FakeTemplateResolver() : super() {
    this._cmpTemplates = MapWrapper.create();
  }
  setTemplate(Type component, Template template) {
    MapWrapper.set(this._cmpTemplates, component, template);
  }
  Template resolve(Type component) {
    var override = MapWrapper.get(this._cmpTemplates, component);
    if (isPresent(override)) {
      return override;
    }
    return super.resolve(component);
  }
}
